#!python3
# Copyright (c) 2021 Food-X Technologies
#
# This file is part of foodx_devops_tools.
#
# You should have received a copy of the MIT License along with
# foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Validation pipeline configuration data."""

import enum
import pathlib
import sys

import click

from ._paths import ConfigurationPathsError, acquire_configuration_paths
from ._version import acquire_version
from .console import report_failure, report_success
from .pipeline_config import (
    ClientsDefinitionError,
    DeploymentsDefinitionError,
    FrameDefinitionsError,
    PipelineConfiguration,
    PipelineConfigurationError,
    PipelineViewError,
    ReleaseStatesDefinitionError,
    SubscriptionsDefinitionError,
    SystemsDefinitionError,
    TenantsDefinitionError,
)


@enum.unique
class ExitState(enum.Enum):
    """Validate configuration script exit states."""

    UNKNOWN = 100
    MISSING_GITREF = 101
    GITREF_PARSE_FAILURE = 102
    BAD_CONFIGURATION_PATHS = 103


@click.command()
@click.version_option(acquire_version())
@click.argument(
    "client_config",
    type=click.Path(dir_okay=True, file_okay=False, path_type=pathlib.Path),
)
@click.argument(
    "system_config",
    type=click.Path(dir_okay=True, file_okay=False, path_type=pathlib.Path),
)
def _main(client_config: pathlib.Path, system_config: pathlib.Path) -> None:
    """Validate pipeline configuration files.

    Exits non-zero if validation fails.

    CONFIGURATION_DIR is the path to the directory where pipeline configuration
    files are located.
    """
    try:
        configuration_paths = acquire_configuration_paths(
            client_config, system_config
        )
        PipelineConfiguration.from_files(configuration_paths)

        report_success("pipeline configuration validated")
    except ConfigurationPathsError as e:
        report_failure(str(e))
        sys.exit(ExitState.BAD_CONFIGURATION_PATHS.value)
    except PipelineConfigurationError as e:
        report_failure(str(e))
        sys.exit(ExitState.MISSING_GITREF.value)
    except (
        PipelineConfigurationError,
        ClientsDefinitionError,
        DeploymentsDefinitionError,
        FrameDefinitionsError,
        ReleaseStatesDefinitionError,
        SubscriptionsDefinitionError,
        SystemsDefinitionError,
        TenantsDefinitionError,
        PipelineViewError,
    ) as e:
        report_failure("Configuration schema error, {0}".format(e))
        sys.exit(ExitState.BAD_CONFIGURATION_PATHS.value)
    except Exception as e:
        report_failure("unknown error, {0}".format(str(e)))
        sys.exit(ExitState.UNKNOWN.value)


def flit_entry() -> None:
    """Flit script entry function for ``validate-configuration`` utility."""
    _main()



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import asyncio
import logging
import pathlib
import sys
import typing

import click

from foodx_devops_tools._paths import (
    ConfigurationPathsError,
    acquire_configuration_paths,
)
from foodx_devops_tools._version import acquire_version
from foodx_devops_tools.pipeline_config import (
    DeploymentContext,
    PipelineConfiguration,
    ReleaseView,
)
from foodx_devops_tools.release_flow import (
    identify_release_id,
    identify_release_state,
)
from foodx_devops_tools.utility import run_command

from ._state import ExitState

log = logging.getLogger(__name__)


class DeploymentConfigurationError(Exception):
    """Problem acquiring deployment configuration."""


def _get_sha() -> str:
    result = run_command(
        ["git", "rev-parse", "HEAD"], text=True, capture_output=True
    )

    if result.returncode != 0:
        raise RuntimeError("Git commit SHA acquisition failed")
    else:
        this_sha = result.stdout[0:10]

    return this_sha


@click.command()
@click.version_option(acquire_version())
@click.argument(
    "client_config",
    type=click.Path(dir_okay=True, file_okay=False, path_type=pathlib.Path),
)
@click.argument(
    "system_config",
    type=click.Path(dir_okay=True, file_okay=False, path_type=pathlib.Path),
)
@click.option(
    "--git-ref",
    default=None,
    help="Git reference to drive release management workflow",
    type=str,
)
@click.option(
    "--pipeline-id",
    default="000+local",
    help="""The pipeline ID. Should be considered mandatory for a CI/CD
 pipeline (optional for developers running locally).

  eg.``$(Build.BuildNumber)`` for Azure DevOps Pipelines.
     ``${GITHUB_RUN_NUMBER}`` for Github Actions.
     ``${CI_PIPELINE_ID}`` for Gitlab-CI.""",
    type=str,
)
def deploy_me(
    client_config: pathlib.Path,
    system_config: pathlib.Path,
    git_ref: typing.Optional[str],
    pipeline_id: str,
) -> None:
    """
    Deploy system resources.

    CLIENT_CONFIG  The client specific configuration directory.
    SYSTEM_CONFIG  The directory containing all non-client related pipeline
                   and deployment configuration.
    """
    try:
        configuration_paths = acquire_configuration_paths(
            client_config, system_config
        )
        this_configuration = PipelineConfiguration.from_files(
            configuration_paths
        )

        if git_ref:
            release_id = identify_release_id(git_ref)
            release_state = identify_release_state(git_ref)
        else:
            raise DeploymentConfigurationError(
                "--git-ref is mandatory (for now)"
            )

        commit_sha = _get_sha()
        base_context = DeploymentContext(
            commit_sha=commit_sha,
            pipeline_id=pipeline_id,
            release_id=release_id,
            release_state=release_state.name,
        )
        pipeline_state = ReleaseView(this_configuration, base_context)
        deployment_iterations = pipeline_state.flatten()

        log.debug(str(deployment_iterations))
    except (ConfigurationPathsError, DeploymentConfigurationError) as e:
        message = str(e)
        log.error(message)
        click.echo(message, err=True)
        sys.exit(ExitState.BAD_DEPLOYMENT_CONFIGURATION.value)
    except asyncio.CancelledError:
        log.debug("Async cancellation exception")
        raise
    except Exception as e:
        log.exception(str(e))
        click.echo(
            "Deployment failed with unexpected error (see log for "
            "details), {0}".format(str(e)),
            err=True,
        )
        sys.exit(ExitState.UNKNOWN_ERROR.value)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Subscriptions deployment configuration I/O."""

import pathlib
import typing

import pydantic

from ._loader import load_configuration

ENTITY_NAME = "subscriptions"


class SubscriptionsDefinitionError(Exception):
    """Problem loading subscription definitions."""


class SingularSubscriptionDefinition(pydantic.BaseModel):
    """Single subscription definition."""

    # Azure DevOps sevice connection name mapping to this subscription
    ado_service_connection: typing.Optional[str]
    # Azure subscription id
    azure_id: str
    # tenant name this subscription maps to
    tenant: str


T = typing.TypeVar("T", bound="SubscriptionsDefinition")

ValueType = typing.Dict[str, SingularSubscriptionDefinition]


class SubscriptionsDefinition(pydantic.BaseModel):
    """Definition of deployment Azure subscriptions."""

    subscriptions: ValueType

    @pydantic.validator(ENTITY_NAME)
    def check_subscriptions(
        cls: pydantic.BaseModel, value: ValueType
    ) -> ValueType:
        """Validate ``subscriptions`` field."""
        if not value:
            raise ValueError("Empty subscription definition prohibited")

        return value


def load_subscriptions(file_path: pathlib.Path) -> SubscriptionsDefinition:
    """
    Load system definitions from file.

    Args:
        file_path: Path to system definitions file.

    Returns:
        System definitions.
    Raises:
        SystemsDefinitionError: If an error occurs loading the file.
    """
    result = load_configuration(
        file_path,
        SubscriptionsDefinition,
        SubscriptionsDefinitionError,
        ENTITY_NAME,
    )

    return typing.cast(SubscriptionsDefinition, result)



# Copyright (c) 2021 Food-X Technologies
#
# This file is part of foodx_devops_tools.
#
# You should have received a copy of the MIT License along with
# foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""NPM CI release flow utility."""

import json
import pathlib
import re

import click

from ._simple_ci_release_id import acquire_release_id

MAX_COMMIT_SHA_LENGTH = 7

OUTPUT_TYPES = [
    "id",
    "package",
]

PACKAGE_EXTENSION = "tgz"


def apply_package_release_id(
    package_file: pathlib.Path, release_id: str
) -> str:
    """
    Apply calculated release id to ``package.json`` file.

    Args:
        package_file: ``package.json`` file to read and modify.
        release_id: release id to apply.

    Returns:
        Package name read from ``package.json`` file.
    """
    with package_file.open(mode="r") as f:
        content = json.load(f)

    content["version"] = release_id

    # write modified json
    with package_file.open(mode="w") as f:
        json.dump(content, f)

    return content["name"]


def normalize_package_name(original_name: str) -> str:
    """
    Convert the raw name from ``package.json`` into the ``npm pack`` style name.

    Args:
        original_name: Name acquired from ``package.json``.

    Returns:
        Normalized name.
    """
    regex = re.compile(r"^(@(?P<group>[0-9a-z\-]+)/)?(?P<package>[0-9a-z\-]+)$")
    result = regex.match(original_name)
    if result and result.group("group"):
        normalized = "{0}-{1}".format(
            result.group("group"), result.group("package")
        )
    elif result:
        normalized = result.group("package")
    else:
        raise RuntimeError(
            "Unable to normalize package name, {0}".format(original_name)
        )

    return normalized


def output_package_name(package_name_json: str, release_id: str) -> None:
    """
    Output to stdout the normalized package file name.

    Args:
        package_name_json: Package name acquired from ``package.json``.
        release_id: Calculated release id.
    """
    package_name_file = normalize_package_name(package_name_json)
    click.echo(
        "{0}-{1}.{2}".format(package_name_file, release_id, PACKAGE_EXTENSION),
        nl=False,
    )


@click.command()
@click.argument(
    "output_type",
    type=click.Choice(OUTPUT_TYPES),
)
@click.argument(
    "package_json_file",
    type=click.Path(file_okay=True, dir_okay=False),
)
@click.argument(
    "ci_ref",
    type=str,
)
@click.argument(
    "commit_sha",
    type=str,
)
def npm_subcommand(
    output_type: str, package_json_file: str, ci_ref: str, commit_sha: str
) -> None:
    """From git reference extract release id for "simple" CI release flow."""
    release_id = acquire_release_id(ci_ref, commit_sha)
    if output_type == "id":
        click.echo(release_id, nl=False)
    elif output_type == "package":
        package_name_json = apply_package_release_id(
            pathlib.Path(package_json_file), release_id
        )

        output_package_name(package_name_json, release_id)
    else:
        raise RuntimeError("Unknown output type, {0}".format(output_type))



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import logging
import pathlib

import pytest

from foodx_devops_tools.pipeline_config import (
    ClientsDefinition,
    DeploymentsDefinition,
    FramesDefinition,
    PipelineConfiguration,
    PipelineConfigurationError,
    PipelineConfigurationPaths,
    ReleaseStatesDefinition,
    SubscriptionsDefinition,
    SystemsDefinition,
    TenantsDefinition,
)
from tests.ci.support.pipeline_config import (
    CLEAN_SPLIT,
    MOCK_PATHS,
    MOCK_RESULTS,
    split_directories,
)

log = logging.getLogger(__name__)

PIPELINE_CONFIG_FILES = {
    "clients.yml",
    "release_states.yml",
    "deployments.yml",
    "frames.yml",
    "subscriptions.yml",
    "systems.yml",
    "tenants.yml",
}


@pytest.fixture()
def mock_loads(mocker):
    def _apply(mock_data: PipelineConfiguration) -> None:
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_clients",
            return_value=ClientsDefinition.parse_obj(
                {"clients": mock_data.clients}
            ),
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_deployments",
            return_value=DeploymentsDefinition.parse_obj(
                {"deployments": mock_data.deployments}
            ),
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_frames",
            return_value=FramesDefinition.parse_obj(
                {"frames": mock_data.frames}
            ),
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_release_states",
            return_value=ReleaseStatesDefinition.parse_obj(
                {"release_states": mock_data.release_states}
            ),
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_subscriptions",
            return_value=SubscriptionsDefinition.parse_obj(
                {"subscriptions": mock_data.subscriptions}
            ),
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_systems",
            return_value=SystemsDefinition.parse_obj(
                {"systems": mock_data.systems}
            ),
        )
        mocker.patch(
            "foodx_devops_tools.pipeline_config.pipeline.load_tenants",
            return_value=TenantsDefinition.parse_obj(
                {"tenants": mock_data.tenants}
            ),
        )

    return _apply


@pytest.fixture()
def mock_results():
    results_copy = MOCK_RESULTS.copy()
    pipeline_config = PipelineConfiguration.parse_obj(results_copy)

    return pipeline_config


def _acquire_configuration_paths(
    client_config: pathlib.Path, system_config: pathlib.Path
) -> PipelineConfigurationPaths:
    """Acquire system, pipeline configuration paths."""
    client_files = [
        x
        for x in client_config.iterdir()
        if x.is_file() and x.name in PIPELINE_CONFIG_FILES
    ]
    system_files = [
        x
        for x in system_config.iterdir()
        if x.is_file() and x.name in PIPELINE_CONFIG_FILES
    ]

    path_arguments = {
        x.name.strip(".yml"): x
        for x in set(client_files).union(set(system_files))
    }
    result = PipelineConfigurationPaths(**path_arguments)
    return result


class TestPipelineConfiguration:
    def test_default(self, mock_loads, mock_results):
        mock_loads(mock_results)
        under_test = PipelineConfiguration.from_files(MOCK_PATHS)

        assert len(under_test.clients) == 2
        assert "c1" in under_test.clients
        assert len(under_test.release_states) == 2
        assert under_test.release_states[0] == "r1"
        assert len(under_test.subscriptions) == 1
        assert (
            under_test.subscriptions["sub1"].ado_service_connection
            == "some-name"
        )
        assert under_test.subscriptions["sub1"].azure_id == "abc123"
        assert len(under_test.systems) == 2
        assert under_test.systems[0] == "sys1"

        assert len(under_test.tenants) == 1
        assert under_test.tenants["t1"].azure_id == "123abc"

    def test_bad_deployment_name_raises(self, mock_loads, mock_results):
        mock_results.deployments = DeploymentsDefinition.parse_obj(
            {
                "deployments": {
                    "badname": {
                        "subscriptions": {
                            "sub1": {
                                "locations": [
                                    {"primary": "l1"},
                                    {"primary": "l2"},
                                ]
                            },
                        },
                    },
                }
            }
        ).deployments
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError, match=r"Bad deployment tuple"
        ):
            PipelineConfiguration.from_files(MOCK_PATHS)

    def test_bad_deployment_clients_raises(self, mock_loads, mock_results):
        mock_results.deployments = DeploymentsDefinition.parse_obj(
            {
                "deployments": {
                    "sys1-bad-r1": {
                        "subscriptions": {
                            "sub1": {
                                "locations": [
                                    {"primary": "l1"},
                                    {"primary": "l2"},
                                ]
                            }
                        },
                    },
                }
            }
        ).deployments
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError, match=r"Bad client in deployment tuple"
        ):
            PipelineConfiguration.from_files(MOCK_PATHS)

    def test_bad_deployment_release_states_raises(
        self, mock_loads, mock_results
    ):
        mock_results.deployments = DeploymentsDefinition.parse_obj(
            {
                "deployments": {
                    "s1-c1-bad": {
                        "subscriptions": {
                            "subname": {
                                "locations": [
                                    {"primary": "l1"},
                                    {"primary": "l2"},
                                ]
                            }
                        },
                    },
                }
            }
        ).deployments
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError,
            match=r"Bad release state in deployment tuple",
        ):
            PipelineConfiguration.from_files(MOCK_PATHS)

    def test_bad_deployment_systems_raises(self, mock_loads, mock_results):
        mock_results.deployments = DeploymentsDefinition.parse_obj(
            {
                "deployments": {
                    "bad-c1-r1": {
                        "subscriptions": {
                            "subname": {
                                "locations": [
                                    {"primary": "l1"},
                                    {"primary": "l2"},
                                ]
                            }
                        },
                    },
                }
            }
        ).deployments
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError, match=r"Bad system in deployment tuple"
        ):
            PipelineConfiguration.from_files(MOCK_PATHS)

    def test_bad_deployment_subscription_raises(self, mock_loads, mock_results):
        mock_results.deployments = DeploymentsDefinition.parse_obj(
            {
                "deployments": {
                    "sys1-c1-r1": {
                        "subscriptions": {
                            "bad_name": {
                                "locations": [
                                    {"primary": "l1"},
                                    {"primary": "l2"},
                                ]
                            }
                        },
                    },
                }
            }
        ).deployments
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError,
            match=r"Bad subscription in deployment",
        ):
            PipelineConfiguration.from_files(MOCK_PATHS)

    def test_bad_client_release_states_raises(self, mock_loads, mock_results):
        mock_results.clients = ClientsDefinition.parse_obj(
            {
                "clients": {
                    "c1": {"release_states": ["bad_state"], "system": "s1"},
                },
            }
        ).clients
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError, match=r"Bad release state in client"
        ):
            PipelineConfiguration.from_files(MOCK_PATHS)

    def test_bad_client_system_raises(self, mock_loads, mock_results):
        mock_results.clients = ClientsDefinition.parse_obj(
            {
                "clients": {
                    "c1": {"release_states": ["r1"], "system": "bad_system"},
                },
            }
        ).clients
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError, match=r"Bad system in client"
        ):
            PipelineConfiguration.from_files(MOCK_PATHS)

    def test_bad_subscription_tenant_raises(self, mock_loads, mock_results):
        mock_results.subscriptions = SubscriptionsDefinition.parse_obj(
            {
                "subscriptions": {
                    "sub1": {
                        "ado_service_connection": "some-name",
                        "azure_id": "abc123",
                        "tenant": "bad_name",
                    },
                },
            }
        ).subscriptions
        mock_loads(mock_results)

        with pytest.raises(
            PipelineConfigurationError, match=r"Bad tenant in subscription"
        ):
            PipelineConfiguration.from_files(MOCK_PATHS)

    def test_load_files(self):
        with split_directories(CLEAN_SPLIT.copy()) as (
            client_config,
            system_config,
        ):
            this_paths = _acquire_configuration_paths(
                client_config, system_config
            )
            # just expect no exceptions, for now.
            PipelineConfiguration.from_files(this_paths)

    def test_load_dict(self):
        under_test = PipelineConfiguration.parse_obj(MOCK_RESULTS.copy())



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import contextlib

import pytest

from foodx_devops_tools.puff.puffignore import BASE_IGNORE, load_puffignore
from tests.ci.support.isolation import simple_isolated_filesystem

MOCK_PUFFIGNORE_CONTENT = """
# a comment
**/some-file.yml
**/ignore-this-dir/*
  
# another comment
**/ignore_another_dir/

"""


@contextlib.contextmanager
def mock_puffignore(content: str):
    with simple_isolated_filesystem() as dir:
        puffignore_path = dir / ".puffignore"
        with open(puffignore_path, mode="w") as f:
            f.write(content)

        yield puffignore_path


@pytest.mark.asyncio
async def test_whitespace_dropped():
    with mock_puffignore(MOCK_PUFFIGNORE_CONTENT) as puffignore_path:
        result = await load_puffignore(puffignore_path)

    expected_result = BASE_IGNORE + [
        "**/some-file.yml",
        "**/ignore-this-dir/*",
        "**/ignore_another_dir/*",
    ]

    assert expected_result == result


@pytest.mark.asyncio
async def test_appended_wildcard():
    """.gitignore convention allows for implicit wildcard terminating
    directory."""
    with mock_puffignore(
        """
**/ignore_another_dir/
"""
    ) as puffignore_path:
        result = await load_puffignore(puffignore_path)

    expected_result = BASE_IGNORE + [
        "**/ignore_another_dir/*",
    ]

    assert expected_result == result


@pytest.mark.asyncio
async def test_empty_loads_clean():
    with mock_puffignore("") as puffignore_path:
        result = await load_puffignore(puffignore_path)

    expected_result = BASE_IGNORE

    assert expected_result == result


@pytest.mark.asyncio
async def test_functionally_empty_loads_clean():
    with mock_puffignore("# just a comment\n") as puffignore_path:
        result = await load_puffignore(puffignore_path)

    expected_result = BASE_IGNORE

    assert expected_result == result


@pytest.mark.asyncio
async def test_nonexistent_file_loads_clean():
    with simple_isolated_filesystem() as dir:
        puffignore_path = dir / ".puffignore"
        result = await load_puffignore(puffignore_path)

    expected_result = BASE_IGNORE

    assert expected_result == result



# Copyright (c) 2021 Food-X Technologies
#
# This file is part of foodx_devops_tools.
#
# You should have received a copy of the MIT License along with
# foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with foodx_devops_tools.
#  If not, see <https://opensource.org/licenses/MIT>.

import pathlib

import pytest

from foodx_devops_tools.pipeline_config import (
    FrameDefinitionsError,
    load_frames,
)


@pytest.fixture
def apply_applications_test(apply_pipeline_config_test):
    def _apply(mock_content: str):
        result = apply_pipeline_config_test(mock_content, load_frames)
        return result

    return _apply


def test_single_default(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          - resource_group: a1_group
            mode: Incremental
        a2:
          - resource_group: a2_group
            mode: Complete
      folder: some/path
"""

    result = apply_applications_test(file_text)

    result_frames = result.frames
    assert len(result_frames.frames) == 1
    assert len(result_frames.frames["f1"].applications["a1"]) == 1
    assert (
        result_frames.frames["f1"].applications["a1"][0].resource_group
        == "a1_group"
    )
    assert len(result_frames.frames["f1"].applications["a2"]) == 1
    assert (
        result_frames.frames["f1"].applications["a2"][0].resource_group
        == "a2_group"
    )
    assert result_frames.frames["f1"].folder == pathlib.Path("some/path")
    assert result_frames.frames["f1"].triggers is None
    assert result_frames.frames["f1"].applications["a2"][0].arm_file is None
    assert result_frames.frames["f1"].applications["a2"][0].puff_file is None
    assert result_frames.triggers is None


def test_arm_optional(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          - resource_group: a1_group
            mode: Incremental
        a2:
          - resource_group: a2_group
            mode: Complete
            arm_file: something.json
      folder: some/path
"""

    result = apply_applications_test(file_text)

    result_frames = result.frames
    assert result_frames.frames["f1"].applications["a2"][
        0
    ].arm_file == pathlib.Path("something.json")


def test_global_path_triggers_optional(apply_applications_test):
    file_text = """---
frames:
  triggers:
    paths:
      - "some/glob/**"
      - "*/stuff/*"
  frames:
    f1:
      applications:
        a1: 
          - resource_group: a1_group
            mode: Incremental
        a2:
          - resource_group: a2_group
            mode: Complete
            arm_file: something.json
      folder: some/path
"""

    result = apply_applications_test(file_text)

    result_frames = result.frames
    assert len(result_frames.triggers.paths) == 2
    assert result_frames.triggers.paths[0] == "some/glob/**"
    assert result_frames.triggers.paths[1] == "*/stuff/*"


def test_frame_path_triggers_optional(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          - resource_group: a1_group
            mode: Incremental
        a2:
          - resource_group: a2_group
            mode: Complete
            arm_file: something.json
      folder: some/path
      triggers:
        paths:
          - "some/glob/**"
          - "*/stuff/*"
"""

    result = apply_applications_test(file_text)

    result_frames = result.frames
    assert len(result_frames.frames["f1"].triggers.paths) == 2
    assert result_frames.frames["f1"].triggers.paths[0] == "some/glob/**"
    assert result_frames.frames["f1"].triggers.paths[1] == "*/stuff/*"


def test_puff_optional(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1: 
          - resource_group: a1_group
            mode: Incremental
        a2:
          - resource_group: a2_group
            mode: Complete
            puff_file: something.yml
      folder: some/path
"""

    result = apply_applications_test(file_text)

    result_frames = result.frames
    assert result_frames.frames["f1"].applications["a2"][
        0
    ].puff_file == pathlib.Path("something.yml")


def test_multiple_sequenced(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1:
          - resource_group: f1a1
            mode: Complete
        a2:
          - resource_group: f1a2
            mode: Incremental
      folder: some/f1-path
    f2:
      applications:
        a3:
          - resource_group: f2a3
            mode: Complete
        a4:
          - resource_group: f2a4
            mode: Incremental
      depends_on:
        - f1
      folder: some/f2-path
"""

    result = apply_applications_test(file_text)

    result_frames = result.frames
    assert len(result_frames.frames) == 2
    assert "f1" in result_frames.frames
    assert "f2" in result_frames.frames
    assert result_frames.frames["f2"].depends_on[0] == "f1"


def test_multiple_unsequenced(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        a1:
          - resource_group: f1a1
            mode: Complete
        a2:
          - resource_group: f1a2
            mode: Complete
      folder: some/f1-path
    f3:
      applications:
        a5:
          - resource_group: f3a5
            mode: Complete
        a6:
          - resource_group: f3a6
            mode: Complete
      folder: some/f3-path
    f2:
      applications:
        a3:
          - resource_group: f2a3
            mode: Complete
        a4:
          - resource_group: f2a4
            mode: Complete
      folder: some/f2-path
"""

    result = apply_applications_test(file_text)

    result_frames = result.frames
    assert len(result_frames.frames) == 3
    assert all([x in result_frames.frames for x in ["f1", "f2", "f3"]])


def test_none_raises(apply_applications_test):
    file_text = """---
"""

    with pytest.raises(
        FrameDefinitionsError, match=r"Error validating frames definition"
    ):
        apply_applications_test(file_text)


def test_empty_raises(apply_applications_test):
    file_text = """---
frames:
"""

    with pytest.raises(
        FrameDefinitionsError, match=r"Error validating frames definition"
    ):
        apply_applications_test(file_text)


def test_bad_dependency_raises(apply_applications_test):
    file_text = """---
frames:
  frames:
    f1:
      applications:
        - a1
        - a2
      folder: some/f1-path
    f2:
      applications:
        - a3
        - a4
      depends_on:
        - bad_value
      folder: some/f2-path
"""

    with pytest.raises(
        FrameDefinitionsError, match=r"Error validating frames definition"
    ):
        apply_applications_test(file_text)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import contextlib
import pathlib

import pytest

from foodx_devops_tools.validate_configuration import _main
from tests.ci.support.click_runner import (  # noqa: F401
    click_runner,
    isolated_filesystem,
)
from tests.ci.support.pipeline_config import (
    CLEAN_SPLIT,
    NOT_SPLIT,
    split_directories,
)

# Due to validation checks in pipeline configuration, each "good" sample must
# be self-consistent with all other samples.
SAMPLES = {
    "clients.yml": {
        "good": """---
clients:
  c1:
    release_states:
      - r1
    system: sys1
""",
        "bad": "",
    },
    "deployments.yml": {
        "good": """---
deployments:
  sys1-c1-r1:
    subscriptions: 
      sub1:
        locations:
          - primary: loc1
          - primary: loc2
""",
        "bad": "",
    },
    "frames.yml": {
        "good": """---
frames:
  frames:
    f1:
      applications:
        a1: 
          - resource_group: a1_group
            mode: Incremental
        a2:
          - resource_group: a2_group
            mode: Complete
            arm_file: something.json
      folder: some/path
      triggers:
        paths:
          - "some/glob/**"
          - "*/stuff/*"
""",
        "bad": "",
    },
    "release_states.yml": {
        "good": """---
release_states:
  - r1
""",
        "bad": "",
    },
    "subscriptions.yml": {
        "good": """---
subscriptions:
  sub1:
    ado_service_connection: some-name
    azure_id: abc123
    tenant: t1
""",
        "bad": "",
    },
    "systems.yml": {
        "good": """---
systems:
  - sys1
""",
        "bad": "",
    },
    "tenants.yml": {
        "good": """---
tenants:
  t1:
    azure_id: abc123
    azure_name: some name
""",
        "bad": "",
    },
}


@contextlib.contextmanager
def mock_files(
    current_file: str,
    current_state: str,
    expected_dir: pathlib.Path,
    mock_runner,
):
    with isolated_filesystem(str(expected_dir), mock_runner):
        for file, content in SAMPLES.items():
            if file != current_file:
                with (expected_dir / file).open(mode="w") as f:
                    f.write(content["good"])
            else:
                with (expected_dir / file).open(mode="w") as f:
                    f.write(content[current_state])

        yield expected_dir


def test_good_exits_clean(click_runner):
    with split_directories(CLEAN_SPLIT.copy()) as (
        client_config,
        system_config,
    ):
        result = click_runner.invoke(
            _main,
            [
                str(client_config),
                str(system_config),
            ],
        )

        if result.exit_code != 0:
            pytest.fail(result.output)


def test_bad_exits_dirty(click_runner):
    expected_dir = pathlib.Path("badpath")
    expected_state = "bad"

    is_bad = list()
    failed_names = list()
    for file, content in SAMPLES.items():
        with mock_files(file, expected_state, expected_dir, click_runner):
            result = click_runner.invoke(_main, [expected_dir])

            if result.exit_code == 0:
                failed_names.append(file)
            is_bad.append(result.exit_code != 0)

    if not all(is_bad):
        pytest.fail(
            "Configuration passed validation when fail expected, "
            "{0}".format(failed_names)
        )



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Azure resource group utilities."""

import asyncio
import json
import logging
import pathlib
import re

from foodx_devops_tools.utility import run_async_command

from .model import AzureSubscriptionConfiguration

log = logging.getLogger(__name__)

AZURE_GROUP_ID_PATTERN = (
    r"^/subscriptions/"
    r"(?P<subscription_id>[0-9a-z\-]+)"
    r"/resourceGroups/"
    r"(?P<group_name>[A-Za-z0-9\-_]+)"
)
AZURE_GROUP_MATCH = re.compile(AZURE_GROUP_ID_PATTERN)


class ResourceGroupError(Exception):
    """Problem occurred with resource group related actions."""


async def check_exists(
    resource_group_name: str, subscription: AzureSubscriptionConfiguration
) -> dict:
    """
    Determine if a resource group exists in the specified subscription.

    Assume authentication has already occurred such that subsequent ``az``
    CLI commands will succeed.

    Args:
        resource_group_name: Name of resource group to find
        subscription: Subscription to search for resource group

    Returns:
        Query result data on success. Empty dictionary otherwise.
    Raises:
        ResourceGroupError: If any error occurs preventing the search from
                            completing.
    """
    try:
        this_command = [
            "az",
            "group",
            "list",
            "--subscription",
            subscription.subscription_id,
        ]
        log.debug("{0}".format(str(this_command)))
        result = await run_async_command(this_command)
        log.debug(
            "resource group existence check stdout, {0}".format(result.out)
        )
        log.debug(
            "resource group existence check stderr, {0}".format(result.error)
        )

        result_data = json.loads(result.out)
        log.debug("{0}".format(str(result_data)))
        group_result = dict()
        for this_group in result_data:
            this_match = AZURE_GROUP_MATCH.match(this_group["id"])
            if (
                this_match
                and this_match.group("group_name") == resource_group_name
            ):
                group_result = this_group.copy()
                log.debug("group exists, {0}".format(str(group_result)))

        return group_result
    except asyncio.CancelledError:
        # should almost always let async cancelled exceptions propagate.
        raise
    except Exception as e:
        raise ResourceGroupError(
            "Problem acquiring resource group data, {0}".format(
                resource_group_name
            )
        ) from e


async def create(
    resource_group_name: str,
    location: str,
    subscription: AzureSubscriptionConfiguration,
) -> None:
    """
    Idempotent creation of an Azure resource group.

    Args:
        resource_group_name: Name of resource group to create.
        location: Azure location in which to create the resource group.
        subscription: Azure subscription identity for the resource group.
    """
    group_data = await check_exists(resource_group_name, subscription)
    if not group_data:
        result = await run_async_command(
            [
                "az",
                "group",
                "create",
                "--resource-group",
                resource_group_name,
                "--location",
                location,
            ]
        )
        log.debug("resource group creation stdout, {0}".format(result.out))
        log.debug("resource group creation stderr, {0}".format(result.error))


async def delete(
    resource_group_name: str,
    subscription: AzureSubscriptionConfiguration,
) -> None:
    """
    Delete an Azure resource group.

    Args:
        resource_group_name: Name of resource group to create.
        subscription: Azure subscription identity for the resource group.
    """
    group_data = await check_exists(resource_group_name, subscription)
    if group_data:
        log.info(
            "deleting resource group, {0}, {1}".format(
                resource_group_name, subscription.subscription_id
            )
        )
        result = await run_async_command(
            [
                "az",
                "group",
                "delete",
                "--resource-group",
                resource_group_name,
                "--yes",
            ]
        )
        log.debug("resource group deletion stdout, {0}".format(result.out))
        log.debug("resource group deletion stderr, {0}".format(result.error))
    else:
        log.info(
            "no delete resource group as it does not exist, {0}, {1}".format(
                resource_group_name, subscription.subscription_id
            )
        )


async def deploy(
    resource_group_name: str,
    arm_template_path: pathlib.Path,
    arm_parameters_path: pathlib.Path,
    location: str,
    subscription: AzureSubscriptionConfiguration,
) -> None:
    """
    Deploy resource to a resource group.

    The resources are defined in ARM template JSON or bicep files, per Azure
    CLI utility.

    Args:
        resource_group_name: Resource group name
        arm_template_path: Path to ARM template deployment file.
        arm_parameters_path: Path to ARM template deployment parameters file.
        location: Azure location in which to deploy resource group.
        subscription: Target subscription/tenant for deployment.
    """
    await create(resource_group_name, location, subscription)
    this_command = [
        "az",
        "deployment",
        "group",
        "create",
        "--resource-group",
        resource_group_name,
        "--parameters",
        "{0}".format(arm_parameters_path),
        "--template-file",
        str(arm_template_path),
    ]
    result = await run_async_command(this_command)
    log.debug("resource group deployment stdout, {0}".format(result.out))
    log.debug("resource group deployment stderr, {0}".format(result.error))



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Manage the structured ``--to`` specifier."""

import dataclasses
import logging
import re
import typing

log = logging.getLogger(__name__)

T = typing.TypeVar("T", bound="StructuredTo")

SPECIFIER_REGEX = (
    r"^(?P<frame>[a-z0-9_\-]+)"
    r"(\."
    r"(?P<application>[a-z0-9_\-]+)"
    r"(\["
    r"(?P<start>[0-9]+)"
    r"(:"
    r"(?P<stop>[0-9]+)"
    r")?"
    r"\])?"
    r")?$"
)


class StructuredToError(Exception):
    """Problem occurred parsing a "--to" specifier."""


@dataclasses.dataclass()
class StructuredTo:
    """Structured deployment specifier."""

    frame: typing.Optional[str] = None
    application: typing.Optional[str] = None
    deployment_range: typing.Optional[range] = None

    @classmethod
    def from_specifier(
        cls: typing.Type[T], to_specified: typing.Optional[str]
    ) -> T:
        """
        Construct a ``StructuredTo`` object from a specifier string.

        Args:
            to_specified: Structured specifier string to be parsed.

        Returns:
            Constructed ``StructuredTo`` object.
        """
        this_object = cls()

        if to_specified:
            this_match = re.match(SPECIFIER_REGEX, to_specified)
            if this_match:
                this_object.frame = this_match.group("frame")

                this_application = this_match.group("application")
                if this_application:
                    this_object.application = this_match.group("application")
                    this_start = this_match.group("start")
                    this_stop = this_match.group("stop")
                    if this_start and (not this_stop):
                        this_object.deployment_range = range(
                            int(this_start), int(this_start) + 1
                        )
                    elif this_start and this_stop:
                        this_object.deployment_range = range(
                            int(this_start), int(this_stop)
                        )
            else:
                # log the error and return the default object
                log.error(
                    "Missing frame from specifier, " "{0}".format(to_specified)
                )
        # else return the default object

        return this_object



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Python clone of node.js puff utility."""

import asyncio
import enum
import logging
import pathlib
import sys

import click
import click_log  # type: ignore

from foodx_devops_tools._version import acquire_version

from .puff import PuffError, run_puff

log = logging.getLogger(__name__)
click_log.basic_config(log)


@enum.unique
class ExitState(enum.Enum):
    """Numeric script exit states."""

    UNKNOWN = 100
    PUFF_FAILED = 101


async def do_run(
    path: pathlib.Path, is_delete_files: bool, is_pretty: bool
) -> None:
    """
    Create or delete puff generated ARM template JSON files.

    Args:
        path: Root directory path to walk.
        is_delete_files: Flag enabling deleting files instead of creating.
        is_pretty: Create nicely formatted JSON for humans.
    """
    await run_puff(path, is_delete_files, is_pretty)


@click.command()
@click_log.simple_verbosity_option(log)
@click.version_option(version=acquire_version())
@click.argument(
    "path",
    type=click.Path(exists=True, file_okay=True, dir_okay=True),
)
@click.option(
    "--delete",
    "-d",
    default=False,
    help="Delete files that were generated.",
    is_flag=True,
    show_default=True,
)
@click.option(
    "--pretty",
    default=False,
    help="Create nicely formatted JSON for humans.",
    is_flag=True,
    show_default=True,
)
def _main(path: str, delete: bool, pretty: bool) -> None:
    """
    Create or delete ARM template parameter files from puff YAML configuration.

    If a directory is specified, all the yaml files in the directory (
    recursively) are assumed to be puff parameter files. If a file is
    specified then just the file is assumed to be a puff parameter file and
    in this case the json files are generated to the parameter files parent
    directory.

    PATH Directory or file path for finding yml files to generate from.
    """
    try:
        asyncio.run(do_run(pathlib.Path(path), delete, pretty))
    except PuffError as e:
        click.echo(str(e), err=True)
        sys.exit(ExitState.PUFF_FAILED.value)
    except asyncio.exceptions.CancelledError:
        log.warning("Exiting due to cancellation")
        # async CancelledError exceptions should usually be re-raised
        # https://docs.python.org/3.8/library/asyncio-exceptions.html#asyncio.CancelledError
        raise
    except Exception as e:
        log.error("Unexpected error, {0}".format(str(e)))
        sys.exit(ExitState.UNKNOWN.value)


def entrypoint() -> None:
    """Flit entry point."""
    _main()



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with foodx_devops_tools.
#  If not, see <https://opensource.org/licenses/MIT>.

import pytest

from foodx_devops_tools.pipeline_config import (
    ReleaseStatesDefinitionError,
    load_release_states,
)


@pytest.fixture
def apply_release_states_test(apply_pipeline_config_test):
    def _apply(mock_content: str):
        result = apply_pipeline_config_test(mock_content, load_release_states)
        return result

    return _apply


def test_single(apply_release_states_test):
    file_text = """
---
release_states:
  - name
"""

    result = apply_release_states_test(file_text)

    assert len(result.release_states) == 1
    assert result.release_states[0] == "name"


def test_multiple(apply_release_states_test):
    file_text = """
---
release_states:
  - name1
  - name2
  - name3
"""

    result = apply_release_states_test(file_text)

    assert len(result.release_states) == 3
    assert result.release_states[0] == "name1"
    assert result.release_states[1] == "name2"
    assert result.release_states[2] == "name3"


def test_duplicate_raises(apply_release_states_test):
    file_text = """
---
release_states:
  - name
  - name
"""

    with pytest.raises(
        ReleaseStatesDefinitionError,
        match=r"Error validating release states definition",
    ):
        apply_release_states_test(file_text)


def test_none_raises(apply_release_states_test):
    file_text = """
---
"""

    with pytest.raises(
        ReleaseStatesDefinitionError,
        match=r"Error validating release states definition",
    ):
        apply_release_states_test(file_text)


def test_empty_list_raises(apply_release_states_test):
    file_text = """
---
release_states:
"""

    with pytest.raises(
        ReleaseStatesDefinitionError,
        match=r"Error validating release states definition",
    ):
        apply_release_states_test(file_text)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import typing
import unittest.mock

import pytest
from asynctest import CoroutineMock


@pytest.fixture()
def mock_async_method(mocker):
    def _apply(
        path_to_mock: str, return_value: typing.Optional[typing.Any] = None
    ):
        async_mock = unittest.mock.AsyncMock(return_value=return_value)
        this_mock = mocker.patch(path_to_mock, side_effect=async_mock)

        return this_mock

    return _apply


@pytest.fixture()
def mock_context(mocker):
    def _apply(path_to_mock: str):
        this_mock = mocker.patch(path_to_mock)
        this_mock.return_value.__aenter__.return_value.write = CoroutineMock()

        return this_mock

    return _apply



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

from foodx_devops_tools.release_flow.npm_ci import (
    acquire_release_id,
    normalize_package_name,
    output_package_name,
)


class TestNormalizePackageName:
    def test_group(self):
        result = normalize_package_name("@this-group/some-name")

        assert result == "this-group-some-name"

    def test_no_group(self):
        result = normalize_package_name("some-name")

        assert result == "some-name"


class TestAcquireReleaseId:
    # Pytest capturing stdout https://docs.pytest.org/en/6.2.x/capture.html
    def test_main_branch(self, mocker):
        mock_arguments = [
            "refs/heads/main",
            "123abc",
        ]
        mocker.patch(
            "foodx_devops_tools.release_flow._simple_ci_release_id.acquire_post_data",
            return_value=("3.1.4", "26"),
        )

        result = acquire_release_id(*mock_arguments)

        assert result == "3.1.4-post.26.123abc"

    def test_long_sha(self, mocker):
        mock_arguments = [
            "refs/heads/main",
            "123abc123abc123abc",
        ]
        mocker.patch(
            "foodx_devops_tools.release_flow._simple_ci_release_id.acquire_post_data",
            return_value=("3.1.4", "26"),
        )

        result = acquire_release_id(*mock_arguments)

        assert result == "3.1.4-post.26.123abc1"

    def test_release_tag(self, capsys, mocker):
        mock_arguments = [
            "refs/tags/31.415.9",
            "123abc",
        ]
        mocker.patch(
            "foodx_devops_tools.release_flow._simple_ci_release_id.acquire_post_data",
            return_value=("3.1.4", "26"),
        )

        result = acquire_release_id(*mock_arguments)

        assert result == "31.415.9"

    def test_semantic_prerelease_tag(self, capsys, mocker):
        """
        Should not be using prerelease tags, but if they are used then
        they are a local release id (essential ignored).
        """
        mock_arguments = [
            "refs/tags/31.415.9-alpha.4",
            "123abc",
        ]
        mocker.patch(
            "foodx_devops_tools.release_flow._simple_ci_release_id.acquire_post_data",
            return_value=("3.1.4", "26"),
        )

        result = acquire_release_id(*mock_arguments)

        assert result == "3.1.4-post.26.123abc"

    def test_dryrun_tag(self, capsys, mocker):
        """Dryrun tag should be preserved as if it were a release."""
        mock_arguments = [
            "refs/tags/31.415.9-dryrun15",
            "123abc",
        ]
        mocker.patch(
            "foodx_devops_tools.release_flow._simple_ci_release_id.acquire_post_data",
            return_value=("3.1.4", "26"),
        )

        result = acquire_release_id(*mock_arguments)

        assert result == "31.415.9-dryrun15"

    def test_feature_branch(self, capsys, mocker):
        mock_arguments = [
            "refs/heads/feature/rcs/some_branch",
            "123abc",
        ]
        mocker.patch(
            "foodx_devops_tools.release_flow._simple_ci_release_id.acquire_post_data",
            return_value=("3.1.4", "26"),
        )

        result = acquire_release_id(*mock_arguments)

        assert result == "3.1.4-post.26.123abc"


class TestOutputPackageName:
    def test_output1(self, capsys):
        arguments = [
            "some-package",
            "3.1.4",
        ]

        output_package_name(*arguments)

        content = capsys.readouterr()
        assert content.out == "some-package-3.1.4.tgz"

    def test_output2(self, capsys):
        arguments = [
            "group-some-package",
            "3.1.4",
        ]

        output_package_name(*arguments)

        content = capsys.readouterr()
        assert content.out == "group-some-package-3.1.4.tgz"



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import copy

import pytest

from foodx_devops_tools.pipeline_config import PipelineConfiguration
from foodx_devops_tools.pipeline_config.views import (
    DeploymentContext,
    DeploymentState,
    DeploymentView,
    PipelineViewError,
    ReleaseView,
    SubscriptionView,
)
from tests.ci.support.pipeline_config import MOCK_RESULTS


@pytest.fixture
def mock_pipeline_config():
    def _apply(mock_data=MOCK_RESULTS.copy()) -> PipelineConfiguration:
        return PipelineConfiguration.parse_obj(mock_data)

    return _apply


MOCK_CONTEXT = DeploymentContext(
    commit_sha="abc123",
    pipeline_id="12345",
    release_id="0.0.0-dev.3",
    release_state="r1",
)


class TestSubscriptionView:
    def test_clean(self, mock_pipeline_config):
        release_view = ReleaseView(mock_pipeline_config(), MOCK_CONTEXT)
        deployment_view = DeploymentView(
            release_view,
            DeploymentState(client="c1", release_state="r1", system="sys1"),
        )
        under_test = SubscriptionView(
            deployment_view,
            "sub1",
        )

    def test_deploy_data_clean(self, mock_pipeline_config):
        release_view = ReleaseView(mock_pipeline_config(), MOCK_CONTEXT)
        deployment_view = DeploymentView(
            release_view,
            DeploymentState(client="c1", release_state="r1", system="sys1"),
        )
        under_test = SubscriptionView(
            deployment_view,
            "sub1",
        )

        result = under_test.deploy_data

        assert [x.ado_service_connection for x in result] == [
            "some-name",
            "some-name",
        ]
        assert {x.location_primary for x in result} == {"l1", "l2"}

    def test_bad_subscription_raises(self, mock_pipeline_config):
        release_view = ReleaseView(mock_pipeline_config(), MOCK_CONTEXT)
        deployment_view = DeploymentView(
            release_view,
            DeploymentState(client="c1", release_state="r1", system="sys1"),
        )
        with pytest.raises(PipelineViewError, match=r"^Bad subscription"):
            SubscriptionView(
                deployment_view,
                "bad_sub",
            )


class TestDeploymentView:
    def test_clean(self, mock_pipeline_config):
        expected_state = DeploymentState(
            client="c1", release_state="r2", system="sys1"
        )
        this_context = copy.deepcopy(MOCK_CONTEXT)
        this_context.release_state = "r2"
        release_view = ReleaseView(mock_pipeline_config(), this_context)
        under_test = DeploymentView(release_view, expected_state)

        assert under_test.release_view == release_view
        assert under_test.deployment_state == expected_state

    def test_bad_deployment_state_raises(self, mock_pipeline_config):
        this_context = copy.deepcopy(MOCK_CONTEXT)
        this_context.release_state = "r2"
        release_view = ReleaseView(mock_pipeline_config(), this_context)
        with pytest.raises(PipelineViewError, match=r"^Bad client"):
            DeploymentView(
                release_view,
                DeploymentState(
                    client="bad_client", release_state="r2", system="sys2"
                ),
            )
        with pytest.raises(PipelineViewError, match=r"^Bad release state"):
            DeploymentView(
                release_view,
                DeploymentState(
                    client="c1", release_state="bad_release", system="sys2"
                ),
            )
        with pytest.raises(PipelineViewError, match=r"^Bad system"):
            DeploymentView(
                release_view,
                DeploymentState(
                    client="c1", release_state="r2", system="bad_system"
                ),
            )

    def test_subscriptions_clean(self, mock_pipeline_config):
        expected_state = DeploymentState(
            client="c1", release_state="r1", system="sys1"
        )
        this_context = copy.deepcopy(MOCK_CONTEXT)
        this_context.release_state = "r2"
        release_view = ReleaseView(mock_pipeline_config(), this_context)
        under_test = DeploymentView(release_view, expected_state)

        assert under_test.release_view == release_view
        assert under_test.deployment_state == expected_state

        assert under_test.subscriptions

    def test_subscriptions_empty(self, mock_pipeline_config):
        this_context = copy.deepcopy(MOCK_CONTEXT)
        this_context.release_state = "r2"
        release_view = ReleaseView(mock_pipeline_config(), this_context)
        under_test = DeploymentView(
            release_view,
            DeploymentState(client="c2", release_state="r2", system="sys2"),
        )

        assert not under_test.subscriptions


class TestReleaseView:
    def test_clean(self, mock_pipeline_config):
        this_context = copy.deepcopy(MOCK_CONTEXT)
        this_context.release_state = "r2"
        under_test = ReleaseView(mock_pipeline_config(), this_context)

        assert under_test.configuration == mock_pipeline_config()
        assert under_test.deployment_context.release_state == "r2"

    def test_bad_release_state_raises(self, mock_pipeline_config):
        this_context = copy.deepcopy(MOCK_CONTEXT)
        this_context.release_state = "r3"
        with pytest.raises(PipelineViewError, match=r"^Bad release state"):
            ReleaseView(mock_pipeline_config(), this_context)

    def test_deployments_clean(self, mock_pipeline_config):
        under_test = ReleaseView(mock_pipeline_config(), MOCK_CONTEXT)

        result = under_test.deployments

        result_tuples = {str(x.deployment_state) for x in result}
        assert result_tuples == {
            "sys1-c1-r1",
        }

    def test_deployments_empty(self, mock_pipeline_config):
        this_context = copy.deepcopy(MOCK_CONTEXT)
        this_context.release_state = "r2"
        under_test = ReleaseView(mock_pipeline_config(), this_context)

        result = under_test.deployments

        result_tuples = {str(x.deployment_state) for x in result}
        assert not result_tuples

    def test_flatten_clean(self, mock_pipeline_config):
        under_test = ReleaseView(mock_pipeline_config(), MOCK_CONTEXT)

        result = under_test.flatten()

        assert len(result) == 2
        assert {x["context"].client for x in result} == {
            "c1",
        }
        assert {x["context"].system for x in result} == {
            "sys1",
        }
        assert {x["data"].location_primary for x in result} == {
            "l1",
            "l2",
        }



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Context sensitive views of pipeline configuration data."""

import copy
import dataclasses
import logging
import typing

from ..deployment import DeploymentState
from .pipeline import PipelineConfiguration

log = logging.getLogger(__name__)


class PipelineViewError(Exception):
    """Problem with view data."""


@dataclasses.dataclass
class DeploymentContext:
    """Deployment context data expected to be applied to resource tags."""

    commit_sha: str
    pipeline_id: str
    release_id: str
    release_state: str

    # these data must be provided at deployment time because they are more
    # deeply context sensitive.
    application_name: typing.Optional[str] = None
    azure_subscription_name: typing.Optional[str] = None
    azure_tenant_name: typing.Optional[str] = None
    client: typing.Optional[str] = None
    frame_name: typing.Optional[str] = None
    system: typing.Optional[str] = None


@dataclasses.dataclass
class DeployDataView:
    """Data critical to a resource deployment."""

    ado_service_connection: str
    azure_subscription_name: str
    azure_tenant_name: str
    deployment_state: str
    location_primary: str
    location_secondary: typing.Optional[str]
    release_state: str


V = typing.TypeVar("V", bound="SubscriptionView")
U = typing.TypeVar("U", bound="DeploymentView")
T = typing.TypeVar("T", bound="ReleaseView")


class SubscriptionView:
    """View of pipeline configuration conditioned by a subscription name."""

    deployment_view: "DeploymentView"
    subscription_name: str

    def __init__(self: V, deployment_view: U, subscription_name: str) -> None:
        """Construct ``SubscriptionView`` object."""
        self.deployment_view = deployment_view
        self.subscription_name = subscription_name

        self._validate_subscription()

    @property
    def deploy_data(self: V) -> typing.List[DeployDataView]:
        """Provide deployment data for each location in this subscription."""
        base_data = (
            self.deployment_view.release_view.configuration.subscriptions[
                self.subscription_name
            ]
        )
        result: typing.List[DeployDataView] = list()
        this_deployment = (
            self.deployment_view.release_view.configuration.deployments[
                str(self.deployment_view.deployment_state)
            ]
        )
        for this_locations in this_deployment.subscriptions[
            self.subscription_name
        ].locations:
            this_data: typing.Dict[str, typing.Any] = {
                "ado_service_connection": base_data.ado_service_connection,
                "azure_subscription_name": self.subscription_name,
                "azure_tenant_name": base_data.tenant,
                "deployment_state": str(self.deployment_view.deployment_state),
                "location_primary": this_locations.primary,
                "location_secondary": None,
                "release_state": self.deployment_view.release_view.deployment_context.release_state,  # noqa E501
            }
            if this_locations.secondary:
                this_data["location_secondary"] = this_locations.secondary

            result.append(DeployDataView(**this_data))

        return result

    def _validate_subscription(self: V) -> None:
        if (
            self.subscription_name
            not in self.deployment_view.release_view.configuration.subscriptions
        ):
            raise PipelineViewError(
                "Bad subscription, {0}".format(str(self.subscription_name))
            )


class DeploymentView:
    """View of pipeline configuration conditioned by deployment state."""

    release_view: "ReleaseView"
    deployment_state: DeploymentState

    def __init__(
        self: U, release_view: T, deployment_state: DeploymentState
    ) -> None:
        """Construct ``DeploymentView`` object."""
        self.release_view = release_view
        self.deployment_state = deployment_state

        self._validate_deployment_state()

    @property
    def subscriptions(self: U) -> typing.List[SubscriptionView]:
        """Provide the subscriptions in this deployment."""
        result: typing.List[SubscriptionView] = list()
        if (
            str(self.deployment_state)
            in self.release_view.configuration.deployments
        ):
            for (
                this_subscription
            ) in self.release_view.configuration.deployments[
                str(self.deployment_state)
            ].subscriptions.keys():
                result.append(SubscriptionView(self, this_subscription))

        return result

    def _validate_deployment_state(self: U) -> None:
        if (
            self.deployment_state.release_state
            not in self.release_view.configuration.release_states
        ):
            raise PipelineViewError(
                "Bad release state, {0}".format(str(self.deployment_state))
            )

        if (
            self.deployment_state.client
            not in self.release_view.configuration.clients.keys()
        ):
            raise PipelineViewError(
                "Bad client, {0}".format(str(self.deployment_state))
            )

        if (
            self.deployment_state.system
            not in self.release_view.configuration.systems
        ):
            raise PipelineViewError(
                "Bad system, {0}".format(str(self.deployment_state))
            )


class ReleaseView:
    """View of pipeline configuration conditioned by release state."""

    configuration: PipelineConfiguration
    deployment_context: DeploymentContext

    def __init__(
        self: T,
        configuration: PipelineConfiguration,
        deployment_context: DeploymentContext,
    ) -> None:
        """Construct ``ReleaseView`` object."""
        self.configuration = configuration
        self.deployment_context = deployment_context

        self._validate_release_state()

    @property
    def deployments(self: T) -> typing.List[DeploymentView]:
        """Provide the deployments in this release."""
        result: typing.List[DeploymentView] = list()
        for this_client in self.configuration.clients.keys():
            for this_system in self.configuration.systems:
                this_state = DeploymentState(
                    client=this_client,
                    release_state=self.deployment_context.release_state,
                    system=this_system,
                )
                if str(this_state) in self.configuration.deployments.keys():
                    result.append(DeploymentView(self, this_state))

        return result

    def flatten(
        self: T,
    ) -> typing.List[dict]:
        """Flatten the nested hierarchy of views into a simple list."""
        result: typing.List[dict] = list()
        for this_deployment in self.deployments:
            for this_subscription in this_deployment.subscriptions:
                for this_deploy_data in this_subscription.deploy_data:
                    updated_context = copy.deepcopy(self.deployment_context)
                    updated_context.azure_subscription_name = (
                        this_subscription.subscription_name
                    )
                    updated_context.azure_tenant_name = (
                        self.configuration.subscriptions[
                            this_subscription.subscription_name
                        ].tenant
                    )
                    updated_context.client = (
                        this_deployment.deployment_state.client
                    )
                    updated_context.system = (
                        this_deployment.deployment_state.system
                    )

                    this_value = {
                        "context": updated_context,
                        "data": this_deploy_data,
                    }
                    result.append(this_value)
        return result

    def _validate_release_state(self: T) -> None:
        if (
            self.deployment_context.release_state
            not in self.configuration.release_states
        ):
            message = "Bad release state, {0}, {1}".format(
                self.deployment_context.release_state,
                str(self.configuration.release_states),
            )
            raise PipelineViewError(message)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import enum


@enum.unique
class ExitState(enum.Enum):
    """Exit state definitions for ``deploy-me`` utility."""

    UNKNOWN_ERROR = 100
    BAD_CLI_ARGUMENTS = 101
    BAD_DEPLOYMENT_CONFIGURATION = 102



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import pathlib

import pytest

from foodx_devops_tools.azure.cloud import AzureSubscriptionConfiguration


def pytest_addoption(parser):
    parser.addoption(
        "--subscription",
        action="store",
        default=None,
        help="Azure Cloud subscription id or name",
    )
    parser.addoption(
        "--parameters",
        action="store",
        default=None,
        help="Path to deployment ARM " "template parameters file",
    )
    parser.addoption(
        "--template",
        action="store",
        default=None,
        help="Path to deployment " "ARM " "template file",
    )


@pytest.fixture
def user_subscription(request) -> AzureSubscriptionConfiguration:
    this_id = request.config.getoption("--subscription")
    this_sub = AzureSubscriptionConfiguration(subscription_id=this_id)
    return this_sub


@pytest.fixture
def template_file(request) -> pathlib.Path:
    return pathlib.Path(request.config.getoption("--template"))


@pytest.fixture
def parameters_file(request) -> pathlib.Path:
    return pathlib.Path(request.config.getoption("--parameters"))



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with foodx_devops_tools.
#  If not, see <https://opensource.org/licenses/MIT>.

import pytest

from foodx_devops_tools.pipeline_config import (
    SubscriptionsDefinitionError,
    load_subscriptions,
)


@pytest.fixture
def apply_subscriptions_test(apply_pipeline_config_test):
    def _apply(mock_content: str):
        result = apply_pipeline_config_test(mock_content, load_subscriptions)
        return result

    return _apply


def test_single(apply_subscriptions_test):
    file_text = """
---
subscriptions:
  name:
    ado_service_connection: some-name
    azure_id: abc123
    tenant: 123abc
"""

    result = apply_subscriptions_test(file_text)

    assert len(result.subscriptions) == 1
    assert "name" in result.subscriptions
    assert result.subscriptions["name"].ado_service_connection == "some-name"
    assert result.subscriptions["name"].azure_id == "abc123"
    assert result.subscriptions["name"].tenant == "123abc"


def test_ado_default(apply_subscriptions_test):
    file_text = """
---
subscriptions:
  name:
    azure_id: abc123
    tenant: tname
"""

    result = apply_subscriptions_test(file_text)

    assert len(result.subscriptions) == 1
    assert "name" in result.subscriptions
    assert result.subscriptions["name"].ado_service_connection is None
    assert result.subscriptions["name"].azure_id == "abc123"
    assert result.subscriptions["name"].tenant == "tname"


def test_multiple(apply_subscriptions_test):
    file_text = """
---
subscriptions:
  name1:
    azure_id: abc123
    tenant: tname1
  name2:
    azure_id: abc1234
    tenant: tname2
  name3:
    azure_id: abc12345
    tenant: tname3
"""

    result = apply_subscriptions_test(file_text)

    assert len(result.subscriptions) == 3
    assert all([x in result.subscriptions for x in ["name1", "name2", "name3"]])
    assert result.subscriptions["name1"].azure_id == "abc123"
    assert result.subscriptions["name3"].azure_id == "abc12345"


def test_none_raises(apply_subscriptions_test):
    file_text = """
---
"""

    with pytest.raises(
        SubscriptionsDefinitionError,
        match=r"Error validating subscriptions definition",
    ):
        apply_subscriptions_test(file_text)


def test_empty_raises(apply_subscriptions_test):
    file_text = """
---
subscriptions:
"""

    with pytest.raises(
        SubscriptionsDefinitionError,
        match=r"Error validating subscriptions definition",
    ):
        apply_subscriptions_test(file_text)



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#
# https://gitlab.com/ci-cd-devops/build_harness/-/blob/main/build_harness/_utility.py

"""General support for build harness implementation."""

import asyncio
import dataclasses
import logging
import subprocess
import typing

log = logging.getLogger(__name__)

CommandArgs = typing.List[str]


class CommandError(Exception):
    """Problem completing an external command run."""


@dataclasses.dataclass()
class CapturedStreams:
    """Encapsulate output and error streams captured from a process."""

    out: str
    error: str


def run_command(
    command: CommandArgs, **kwargs: typing.Any
) -> subprocess.CompletedProcess:
    """
    Run a system command using ``subprocess.run``.

    Args:
        command: List of command and arguments.
        **kwargs: Optional arguments passed through to ``subprocess.run``.

    Returns:
        Subprocess results.
    """
    log.debug("command to run, {0}".format(str(command)))
    log.debug("command arguments, {0}".format(str(kwargs)))
    result = subprocess.run(command, **kwargs)

    return result


async def run_async_command(command: CommandArgs) -> CapturedStreams:
    """
    Run an external command asynchronously.

    Args:
        command: Command to be executed.

    Returns:
        Any output or error streams captured from the process.
    """
    log.debug("command to run, {0}".format(str(command)))
    this_process = await asyncio.create_subprocess_exec(
        *command, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )

    stdout, stderr = await this_process.communicate()
    result = CapturedStreams(
        out=stdout.decode("utf-8"), error=stderr.decode("utf-8")
    )
    if this_process.returncode != 0:
        raise CommandError(
            "External command run did not exit cleanly, {0}".format(
                result.error
            )
        )
    return result



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Tenants deployment configuration I/O."""

import pathlib
import typing

import pydantic

from ._loader import load_configuration

ENTITY_NAME = "tenants"


class TenantsDefinitionError(Exception):
    """Problem loading tenant definitions."""


class SingularTenantDefinition(pydantic.BaseModel):
    """Single tenant definition."""

    azure_id: str
    azure_name: typing.Optional[str]


T = typing.TypeVar("T", bound="TenantsDefinition")

ValueType = typing.Dict[str, SingularTenantDefinition]


class TenantsDefinition(pydantic.BaseModel):
    """Definition of deployment Azure tenants."""

    tenants: ValueType

    @pydantic.validator(ENTITY_NAME)
    def check_tenants(cls: pydantic.BaseModel, value: ValueType) -> ValueType:
        """Validate ``tenants`` field."""
        if not value:
            raise ValueError("Empty tenant definition prohibited")

        return value


def load_tenants(file_path: pathlib.Path) -> TenantsDefinition:
    """
    Load system definitions from file.

    Args:
        file_path: Path to system definitions file.

    Returns:
        System definitions.
    Raises:
        SystemsDefinitionError: If an error occurs loading the file.
    """
    result = load_configuration(
        file_path,
        TenantsDefinition,
        TenantsDefinitionError,
        ENTITY_NAME,
    )

    return typing.cast(TenantsDefinition, result)



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.
import asyncio
import re

import pytest

from foodx_devops_tools.azure.cloud import (
    AzureSubscriptionConfiguration,
    deploy_resource_group,
)
from foodx_devops_tools.azure.cloud.exceptions import ResourceGroupError
from foodx_devops_tools.azure.cloud.resource_group import AZURE_GROUP_ID_PATTERN
from foodx_devops_tools.azure.cloud.resource_group import (
    check_exists as resource_group_exists,
)
from foodx_devops_tools.azure.cloud.resource_group import (
    create as create_resource_group,
)
from foodx_devops_tools.utility import CapturedStreams
from tests.ci.support.asyncio import mock_async_method  # noqa: F401

MOCK_SUBSCRIPTION = AzureSubscriptionConfiguration(
    subscription_id="123-abc", tenant_id="abc-123"
)

MOCKING_PATHS = {
    "group_exists": "foodx_devops_tools.azure.cloud.resource_group"
    ".check_exists",
    "json_loads": "foodx_devops_tools.azure.cloud.resource_group.json.loads",
    "run_async": "foodx_devops_tools.azure.cloud.resource_group"
    ".run_async_command",
}


class TestResourceGroupIdRegex:
    def test_clean(self):
        expected = {
            "subscription_id": "9aaa123-c1ba-4e79-a9af-dbexamplecb",
            "group_name": "some-name",
        }
        data = "/subscriptions/{0}/resourceGroups/{1}".format(
            expected["subscription_id"], expected["group_name"]
        )

        result = re.match(AZURE_GROUP_ID_PATTERN, data)

        assert result
        assert result.group("subscription_id") == expected["subscription_id"]
        assert result.group("group_name") == expected["group_name"]


class TestResourceGroupExists:
    MOCK_SUBSCRIPTION = AzureSubscriptionConfiguration(
        subscription_id="123-abc", tenant_id="abc-123"
    )
    MOCK_RETURN = CapturedStreams(
        out="""[
  {
    "id": "/subscriptions/9aaa123-c1ba-4e79-a9af-dbexamplecb/resourceGroups/some-name",
    "location": "canadacentral",
    "managedBy": null,
    "name": "some-name",
    "properties": {
      "provisioningState": "Succeeded"
    },
    "tags": null,
    "type": "Microsoft.Resources/resourceGroups"
  }
]
""",
        error="",
    )

    @pytest.mark.asyncio
    async def test_clean(self, mock_async_method):
        mock_async_method(
            MOCKING_PATHS["run_async"],
            return_value=self.MOCK_RETURN,
        )

        assert not await resource_group_exists("notagroup", MOCK_SUBSCRIPTION)
        result = await resource_group_exists("some-name", MOCK_SUBSCRIPTION)

        assert result == {
            "id": "/subscriptions/9aaa123-c1ba-4e79-a9af-dbexamplecb/resourceGroups/some-name",
            "location": "canadacentral",
            "managedBy": None,
            "name": "some-name",
            "properties": {"provisioningState": "Succeeded"},
            "tags": None,
            "type": "Microsoft.Resources/resourceGroups",
        }

    @pytest.mark.asyncio
    async def test_raises(self, mock_async_method, mocker):
        mock_async_method(
            MOCKING_PATHS["run_async"],
            return_value=self.MOCK_RETURN,
        )
        mocker.patch(
            MOCKING_PATHS["json_loads"],
            side_effect=RuntimeError("some error"),
        )

        with pytest.raises(
            ResourceGroupError,
            match=r"^Problem acquiring resource group data",
        ):
            await resource_group_exists("some-name", MOCK_SUBSCRIPTION)

    @pytest.mark.asyncio
    async def test_async_cancelled_raises(self, mock_async_method, mocker):
        mock_async_method(
            MOCKING_PATHS["run_async"],
            return_value=self.MOCK_RETURN,
        )
        mocker.patch(
            MOCKING_PATHS["json_loads"],
            side_effect=asyncio.CancelledError(),
        )

        with pytest.raises(asyncio.CancelledError):
            await resource_group_exists("some-name", MOCK_SUBSCRIPTION)


class TestCreateResourceGroup:
    MOCK_RETURN = CapturedStreams(out="good run", error="")
    MOCK_GROUP_DATA = {
        "id": "/subscriptions/9aaa123-c1ba-4e79-a9af-dbexamplecb/resourceGroups/some-name",
        "location": "canadacentral",
        "managedBy": None,
        "name": "some-name",
        "properties": {"provisioningState": "Succeeded"},
        "tags": None,
        "type": "Microsoft.Resources/resourceGroups",
    }

    @pytest.mark.asyncio
    async def test_not_exists_created(self, mock_async_method, mocker):
        mock_run = mock_async_method(
            MOCKING_PATHS["run_async"],
            return_value=self.MOCK_RETURN,
        )
        mock_async_method(
            MOCKING_PATHS["group_exists"],
            return_value=None,
        )

        await create_resource_group(
            "some-name", "canadacentral", MOCK_SUBSCRIPTION
        )

        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        "az",
                        "group",
                        "create",
                        "--resource-group",
                        "some-name",
                        "--location",
                        "canadacentral",
                    ]
                )
            ]
        )

    @pytest.mark.asyncio
    async def test_exists_do_nothing(self, mock_async_method):
        mock_run = mock_async_method(
            MOCKING_PATHS["run_async"],
            return_value=self.MOCK_RETURN,
        )
        mock_async_method(
            MOCKING_PATHS["group_exists"],
            return_value=self.MOCK_GROUP_DATA.copy(),
        )

        await create_resource_group(
            "some-name", "canadacentral", MOCK_SUBSCRIPTION
        )

        mock_run.assert_not_called()


# class TestDeployResourceGroup:
#     MOCK_SUBSCRIPTION = AzureSubscriptionConfiguration(
#         subscription_id="123-abc", tenant_id="abc-123")
#     MOCK_RETURN=CapturedStreams(out="good run", error="")
#     @pytest.mark.asyncio
#     async def test_clean(self,mock_async_method):
#         mock_run=mock_async_method("foodx_devops_tools.cli.run_async_command",
#                                    return_value=self.MOCK_RETURN)
#         await deploy_resource_group(,,self.MOCK_SUBSCRIPTION)
#         assert False



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.
import pydantic
import pytest

from foodx_devops_tools.puff._puff_parameters import (
    PuffEnvironment,
    PuffEnvironments,
    PuffParameterModel,
    PuffRegion,
    PuffRegions,
    PuffService,
    PuffServices,
)


class TestPuffRegion:
    def test_clean(self):
        under_test = PuffRegion.parse_obj(
            {
                "r1": {"n1": "v1", "n2": "v2"},
            }
        )

        assert under_test["r1"] == {"n1": "v1", "n2": "v2"}
        assert "r1" in under_test

    def test_multiple_names_raises(self):
        with pytest.raises(
            pydantic.ValidationError,
            match=r"must only specify a single name in a region entry",
        ):
            PuffRegion.parse_obj(
                {
                    "r1": {"n1": "v1", "n2": "v2"},
                    "r2": {"n2": "v3"},
                }
            )


class TestPuffRegions:
    def test_clean(self):
        under_test = PuffRegions.parse_obj(
            [
                {
                    "r1": {"n1": "v1", "n2": "v2"},
                },
                {
                    "r2": {"n2": "v3"},
                },
            ]
        )

        assert under_test[0]["r1"] == {"n1": "v1", "n2": "v2"}
        assert under_test[1]["r2"] == {"n2": "v3"}

    def test_empty_region(self):
        under_test = PuffRegions.parse_obj(
            [
                {
                    "r1": {},
                },
                {
                    "r2": {"n2": "v3"},
                },
            ]
        )

        assert "r1" in under_test[0]
        assert under_test[1]["r2"] == {"n2": "v3"}

    def test_none_region(self):
        under_test = PuffRegions.parse_obj(
            [
                {
                    "r1": None,
                },
                {
                    "r2": {"n2": "v3"},
                },
            ]
        )

        assert "r1" in under_test[0]
        assert under_test[1]["r2"] == {"n2": "v3"}

    def test_multiple_names_raises(self):
        with pytest.raises(
            pydantic.ValidationError,
            match=r"must only specify a single name in a region entry",
        ):
            PuffRegions.parse_obj(
                [
                    {
                        "r1": {"n1": "v1", "n2": "v2"},
                    },
                    {
                        "r1": {"n1": "v1", "n2": "v2"},
                        "r2": {"n2": "v3"},
                    },
                ]
            )


class TestPuffEnvironment:
    def test_clean_regions_present(self):
        under_test = PuffEnvironment.parse_obj(
            {
                "n1": "v1",
                "regions": [
                    {
                        "r1": {
                            "n2": "v2",
                        },
                    },
                ],
            }
        )
        assert under_test["n1"] == "v1"
        assert under_test["regions"][0]["r1"]["n2"] == "v2"
        assert "n1" in under_test

    def test_clean_regions_absent(self):
        under_test = PuffEnvironment.parse_obj(
            {
                "n1": "v1",
            }
        )
        assert under_test["n1"] == "v1"
        assert "regions" not in under_test

    def test_bad_regions_raises(self):
        with pytest.raises(pydantic.ValidationError):
            PuffEnvironment.parse_obj(
                {
                    "n1": "v1",
                    # regions is not a list
                    "regions": {
                        "n2": "v2",
                    },
                }
            )


class TestPuffEnvironments:
    def test_clean(self):
        under_test = PuffEnvironments.parse_obj(
            {
                "e1": {
                    "n1": "v1",
                    "regions": [
                        {
                            "r1": {
                                "n2": "v2",
                            },
                        }
                    ],
                },
                "e2": {
                    "regions": [
                        {
                            "r1": {
                                "n2": "v3",
                            },
                        }
                    ]
                },
            }
        )

        assert under_test["e1"]["n1"] == "v1"
        assert "n1" not in under_test["e2"]
        assert "e1" in under_test

    def test_empty_environment(self):
        under_test = PuffEnvironments.parse_obj(
            {
                "e1": dict(),
                "e2": {
                    "regions": [
                        {
                            "r1": {
                                "n2": "v3",
                            },
                        }
                    ]
                },
            }
        )

        assert "n1" not in under_test["e2"]
        assert "e1" in under_test

    def test_none_environment(self):
        under_test = PuffEnvironments.parse_obj(
            {
                "e1": None,
                "e2": {
                    "regions": [
                        {
                            "r1": {
                                "n2": "v3",
                            },
                        }
                    ]
                },
            }
        )

        assert "n1" not in under_test["e2"]
        assert "e1" in under_test


class TestPuffService:
    def test_clean(self):
        under_test = PuffService.parse_obj(
            {
                "n1": "v1",
                "environments": {
                    "e1": {
                        "n1": "v2",
                        "regions": {
                            "n2": "v3",
                        },
                    },
                },
            }
        )

        assert under_test["n1"] == "v1"
        assert under_test["environments"]["e1"]["n1"] == "v2"
        assert "environments" in under_test


class TestPuffServices:
    def test_clean(self):
        under_test = PuffServices.parse_obj(
            {
                "s1": {
                    "n1": "v1",
                    "environments": {
                        "e1": {
                            "n1": "v3",
                            "regions": {
                                "n2": "v2",
                            },
                        },
                    },
                },
                "s2": {
                    "n1": "v5",
                    "environments": {
                        "e2": {
                            "regions": {
                                "n2": "v3",
                            },
                        },
                    },
                },
            }
        )

        assert under_test["s1"]["n1"] == "v1"
        assert under_test["s2"]["n1"] == "v5"
        assert "s1" in under_test


class TestPuffParameterModel:
    def test_clean(self):
        under_test = PuffParameterModel.parse_obj(
            {
                "name": "some-name",
                "k1": "v1",
                "environments": {
                    "e1": {
                        "n1": "v1",
                        "regions": [
                            {
                                "r1": {
                                    "n2": "v2",
                                },
                            }
                        ],
                    },
                    "e2": {
                        "regions": [
                            {
                                "r1": {
                                    "n2": "v3",
                                },
                            }
                        ]
                    },
                },
                "services": {
                    "s1": {
                        "n1": "v1",
                        "environments": {
                            "e1": {
                                "n1": "v3",
                                "regions": [
                                    {
                                        "r1": {
                                            "n2": "v2",
                                        },
                                    }
                                ],
                            },
                        },
                    },
                    "s2": {
                        "n1": "v5",
                        "environments": {
                            "e2": {
                                "regions": [
                                    {
                                        "r1": {
                                            "n2": "v3",
                                        },
                                    }
                                ]
                            },
                        },
                    },
                },
            },
        )

    def test_empty_environments(self):
        under_test = PuffParameterModel.parse_obj(
            {
                "name": "some-name",
                "environments": dict(),
            },
        )

        assert "environments" in under_test



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import contextlib
import pathlib
import tempfile

import ruamel.yaml

from foodx_devops_tools.pipeline_config import PipelineConfigurationPaths

CLEAN_SPLIT = {
    "client": {
        "clients",
        "deployments",
        "frames",
    },
    "system": {
        "release_states",
        "subscriptions",
        "systems",
        "tenants",
    },
}
NOT_SPLIT = {
    "client": {},
    "system": {
        "release_states",
        "subscriptions",
        "systems",
        "tenants",
        "clients",
        "deployments",
        "frames",
    },
}

MOCK_RESULTS = {
    "clients": {
        "c1": {
            "release_states": [
                "r1",
                "r2",
            ],
            "system": "sys1",
        },
        "c2": {
            "release_states": [
                "r2",
            ],
            "system": "sys2",
        },
    },
    "deployments": {
        "sys1-c1-r1": {
            "subscriptions": {
                "sub1": {
                    "locations": [{"primary": "l1"}, {"primary": "l2"}],
                },
            },
        },
    },
    "frames": {
        "frames": {
            "f1": {
                "applications": {
                    "a1": [
                        {
                            "resource_group": "a1_group",
                            "mode": "Incremental",
                        },
                    ],
                },
                "folder": "some/path",
            },
        },
    },
    "release_states": ["r1", "r2"],
    "subscriptions": {
        "sub1": {
            "ado_service_connection": "some-name",
            "azure_id": "abc123",
            "tenant": "t1",
        },
    },
    "systems": ["sys1", "sys2"],
    "tenants": {"t1": {"azure_id": "123abc"}},
}

MOCK_PATHS = PipelineConfigurationPaths(
    **{
        "clients": pathlib.Path("client/path"),
        "release_states": pathlib.Path("release_state/path"),
        "deployments": pathlib.Path("deployment/path"),
        "frames": pathlib.Path("frame/path"),
        "subscriptions": pathlib.Path("subscription/path"),
        "systems": pathlib.Path("system/path"),
        "tenants": pathlib.Path("tenant/path"),
    }
)


@contextlib.contextmanager
def split_directories(split: dict):
    """Generate configuration files split across two directories."""
    yaml = ruamel.yaml.YAML(typ="safe")
    with tempfile.TemporaryDirectory() as dir1:
        pd1 = pathlib.Path(dir1)
        with tempfile.TemporaryDirectory() as dir2:
            pd2 = pathlib.Path(dir2)
            paths = {
                "client": pd1,
                "system": pd2,
            }
            for key, files in split.items():
                for this_file in files:
                    file_path = paths[key] / "{0}.yml".format(this_file)
                    with file_path.open(mode="w") as f:
                        yaml.dump({this_file: MOCK_RESULTS[this_file]}, f)

            yield pd1, pd2



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import click

from .._version import acquire_version
from .azure_cd import azure_subcommand
from .npm_ci import npm_subcommand


@click.group()
@click.version_option(version=acquire_version())
def release_flow() -> None:
    """Release flow command group."""


release_flow.add_command(azure_subcommand, name="azure")
release_flow.add_command(npm_subcommand, name="npm")



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Run resource group tests against an actual Azure Cloud subscription."""

import json
import logging
import uuid

import pytest

from foodx_devops_tools.azure.cloud import deploy_resource_group
from foodx_devops_tools.azure.cloud.resource import (
    check_exists as check_resource_exists,
)
from foodx_devops_tools.azure.cloud.resource_group import (
    check_exists as check_resource_group_exists,
)

from .support import (  # noqa: F401
    clean_teardown,
    prexisting_resource_group,
    unique_group_name,
)

log = logging.getLogger(__name__)


class TestDeployResourceGroup:
    @pytest.mark.asyncio
    async def test_existing_resource_group_clean(
        self,
        prexisting_resource_group,
        unique_group_name,
        user_subscription,
        template_file,
        parameters_file,
    ):
        expected_group = unique_group_name
        expected_resource = str(uuid.uuid4()).replace("-", "")[0:20]
        expected_location = "eastus"
        with parameters_file.open(mode="r") as f:
            data = json.load(f)
        data["parameters"]["storageAccountName"]["value"] = expected_resource
        if isinstance(data["parameters"]["location"], str):
            data["parameters"]["location"] = {"value": expected_location}
        else:
            data["parameters"]["location"]["value"] = expected_location
        with parameters_file.open(mode="w") as f:
            json.dump(data, f)

        async with prexisting_resource_group(expected_group):
            await deploy_resource_group(
                expected_group,
                template_file,
                parameters_file,
                expected_location,
                user_subscription,
            )

            await self.do_check(
                expected_group, expected_resource, user_subscription
            )

    async def do_check(self, group_name: str, resource_name: str, subscription):
        assert await check_resource_group_exists(group_name, subscription)
        assert await check_resource_exists(resource_name, subscription)

    @pytest.mark.asyncio
    async def test_nonexisting_resource_group_clean(
        self,
        clean_teardown,
        unique_group_name,
        user_subscription,
        template_file,
        parameters_file,
    ):
        expected_group = unique_group_name
        expected_resource = str(uuid.uuid4()).replace("-", "")[0:20]
        expected_location = "eastus"

        with parameters_file.open(mode="r") as f:
            data = json.load(f)
        data["parameters"]["storageAccountName"]["value"] = expected_resource
        if isinstance(data["parameters"]["location"], str):
            data["parameters"]["location"] = {"value": expected_location}
        else:
            data["parameters"]["location"]["value"] = expected_location
        with parameters_file.open(mode="w") as f:
            json.dump(data, f)

        async with clean_teardown(expected_group):
            await deploy_resource_group(
                expected_group,
                template_file,
                parameters_file,
                expected_location,
                user_subscription,
            )

            await self.do_check(
                expected_group, expected_resource, user_subscription
            )



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import asyncio
import json

import pytest

from foodx_devops_tools.azure.cloud.resource import (
    AzureSubscriptionConfiguration,
    ResourceError,
    check_exists,
    list_resources,
)
from foodx_devops_tools.utility import CapturedStreams
from tests.ci.support.asyncio import mock_async_method  # noqa: F401

MOCKING_PATHS = {
    "json_loads": "foodx_devops_tools.azure.cloud.resource.json.loads",
    "run_async": "foodx_devops_tools.azure.cloud.resource" ".run_async_command",
    "list_resources": "foodx_devops_tools.azure.cloud.resource"
    ".list_resources",
}
MOCK_SUBSCRIPTION = AzureSubscriptionConfiguration(
    subscription_id="123-abc", tenant_id="abc-123"
)

MOCK_DATA = [
    {
        "id": "some-id",
        "name": "some-name",
    },
    {
        "id": "another-id",
        "name": "other-name",
    },
]
MOCK_RETURN = CapturedStreams(out=json.dumps(MOCK_DATA), error="")


class TestListResources:
    @pytest.mark.asyncio
    async def test_name_clean(self, mock_async_method):
        mock_async_method(
            MOCKING_PATHS["run_async"],
            return_value=MOCK_RETURN,
        )

        result = await list_resources(MOCK_SUBSCRIPTION, name="resource-name")

        assert result == MOCK_DATA

    @pytest.mark.asyncio
    async def test_default(self, mock_async_method):
        mock_async_method(
            MOCKING_PATHS["run_async"],
            return_value=list(),
        )

        with pytest.raises(
            ResourceError,
            match=r"^Must specify at least one of name, location or tag",
        ):
            await list_resources(MOCK_SUBSCRIPTION)


class TestCheckExists:
    MOCK_MULTIPLE_DATA = [
        {
            "id": "some-id",
            "name": "some-name",
        },
        {
            "id": "other-id",
            "name": "some-name",
        },
        {
            "id": "another-id",
            "name": "other-name",
        },
    ]
    MOCK_MULTIPLE_RETURN = CapturedStreams(
        out=json.dumps(MOCK_MULTIPLE_DATA), error=""
    )

    @pytest.mark.asyncio
    async def test_resource_found(self, mock_async_method):
        mock_async_method(
            MOCKING_PATHS["run_async"],
            return_value=MOCK_RETURN,
        )

        result = await check_exists("some-name", MOCK_SUBSCRIPTION)

        assert len(result) == 1
        assert result[0]["name"] == "some-name"

    @pytest.mark.asyncio
    async def test_multiple_resources_found(self, mock_async_method):
        mock_async_method(
            MOCKING_PATHS["run_async"],
            return_value=self.MOCK_MULTIPLE_RETURN,
        )

        result = await check_exists("some-name", MOCK_SUBSCRIPTION)
        assert len(result) == 2

    @pytest.mark.asyncio
    async def test_resource_not_found(self, mock_async_method):
        mock_async_method(
            MOCKING_PATHS["run_async"],
            return_value=MOCK_RETURN,
        )

        result = await check_exists("notvalid", MOCK_SUBSCRIPTION)
        assert not result

    @pytest.mark.asyncio
    async def test_async_cancelled_raises(self, mock_async_method, mocker):
        mock_async_method(
            MOCKING_PATHS["run_async"],
            return_value=MOCK_RETURN,
        )
        mocker.patch(
            MOCKING_PATHS["list_resources"],
            side_effect=asyncio.CancelledError(),
        )

        with pytest.raises(asyncio.CancelledError):
            await check_exists("some-name", MOCK_SUBSCRIPTION)



# Copyright (c) 2021 Food-X Technologies
#
# This file is part of foodx_devops_tools.
#
# You should have received a copy of the MIT License along with
# foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""File I/O for pipeline configuration metadata."""

from .clients import (  # noqa: F401
    ClientsDefinition,
    ClientsDefinitionError,
    load_clients,
)
from .deployments import (  # noqa: F401
    DeploymentsDefinition,
    DeploymentsDefinitionError,
    load_deployments,
)
from .frames import (  # noqa: F401
    FrameDefinitionsError,
    FramesDefinition,
    load_frames,
)
from .pipeline import (  # noqa: F401
    PipelineConfiguration,
    PipelineConfigurationError,
    PipelineConfigurationPaths,
)
from .release_states import (  # noqa: F401
    ReleaseStatesDefinition,
    ReleaseStatesDefinitionError,
    load_release_states,
)
from .subscriptions import (  # noqa: F401
    SubscriptionsDefinition,
    SubscriptionsDefinitionError,
    load_subscriptions,
)
from .systems import (  # noqa: F401
    SystemsDefinition,
    SystemsDefinitionError,
    load_systems,
)
from .tenants import (  # noqa: F401
    TenantsDefinition,
    TenantsDefinitionError,
    load_tenants,
)
from .views import (  # noqa: F401
    DeploymentContext,
    PipelineViewError,
    ReleaseView,
)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['foodx_devops_tools',
 'foodx_devops_tools.azure',
 'foodx_devops_tools.azure.cloud',
 'foodx_devops_tools.deploy_me',
 'foodx_devops_tools.pipeline_config',
 'foodx_devops_tools.puff',
 'foodx_devops_tools.release_flow']

package_data = \
{'': ['*']}

install_requires = \
['aiofiles == 0.7.0',
 'azure-devops == 6.0.0b4',
 'click == 8.0.1',
 'click-log == 0.3.2',
 'deepmerge == 0.3.0',
 'pydantic == 1.8.2',
 'ruamel.yaml == 0.17.9']

extras_require = \
{'dev': ['build_harness == 0.1.1', 'pre_commit == 2.7.1'],
 'doc': ['sphinx == 3.2.1', 'sphinx_rtd_theme == 0.5.0'],
 'test': ['asynctest == 0.13.0',
          'pytest == 6.1.1',
          'pytest-cov == 2.10.1',
          'pytest-mock == 2.0.0']}

entry_points = \
{'console_scripts': ['deploy-me = '
                     'foodx_devops_tools.deploy_me_entry:flit_entry',
                     'foodx-release-flow = '
                     'foodx_devops_tools.release_flow_entry:flit_entry',
                     'puff = foodx_devops_tools.puff_utility:entrypoint',
                     'validate-configuration = '
                     'foodx_devops_tools.validate_configuration:flit_entry']}

setup(name='foodx_devops_tools',
      version='0.0.0.post20',
      description='Foodx DevOps pipeline utilities.',
      author='FoodX Technologies',
      author_email='support@foodxtech.com',
      url='https://github.com/Food-X-Technologies/foodx_devops_tools',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.8',
     )



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import enum
import logging

import pytest

from foodx_devops_tools.deploy_me._main import (
    ConfigurationPathsError,
    _get_sha,
    acquire_configuration_paths,
)
from foodx_devops_tools.deploy_me_entry import deploy_me
from foodx_devops_tools.pipeline_config import PipelineConfigurationPaths
from tests.ci.support.click_runner import click_runner  # noqa: F401
from tests.ci.support.pipeline_config import CLEAN_SPLIT, split_directories


class TestAcquireConfigurationPaths:
    def test_clean(self):
        with split_directories(CLEAN_SPLIT.copy()) as (
            client_config,
            system_config,
        ):
            result = acquire_configuration_paths(client_config, system_config)

            assert result == PipelineConfigurationPaths(
                clients=client_config / "clients.yml",
                deployments=client_config / "deployments.yml",
                frames=client_config / "frames.yml",
                release_states=system_config / "release_states.yml",
                subscriptions=system_config / "subscriptions.yml",
                systems=system_config / "systems.yml",
                tenants=system_config / "tenants.yml",
            )

    def test_duplicates_raises(self):
        split = {
            # duplicates between client and system dirs
            "client": {
                "clients",
                "deployments",
                "frames",
                "release_states",
            },
            "system": {
                "release_states",
                "subscriptions",
                "systems",
                "tenants",
            },
        }
        with split_directories(split) as (
            client_config,
            system_config,
        ), pytest.raises(
            ConfigurationPathsError,
            match=r"^Duplicate files between directories",
        ):
            acquire_configuration_paths(client_config, system_config)


@pytest.fixture()
def mock_gitsha_result(mocker):
    def _apply(returncode: int, output: str = ""):
        mock_result = mocker.MagicMock()
        mock_result.returncode = returncode
        mock_result.stdout = output
        this_mock = mocker.patch(
            "foodx_devops_tools.deploy_me._main.run_command",
            return_value=mock_result,
        )
        return this_mock

    return _apply


class TestGetSha:
    def test_clean(self, mock_gitsha_result):
        mock_gitsha_result(0, output="abc123")
        result = _get_sha()

        assert result == "abc123"

    def test_error_raises(self, mock_gitsha_result):
        mock_gitsha_result(1)
        with pytest.raises(RuntimeError):
            _get_sha()


class TestDeployMe:
    class MockReleaseState(enum.Enum):
        r1 = enum.auto()

    def test_default(self, click_runner, caplog, mock_gitsha_result, mocker):
        mock_gitsha_result(0)
        mocker.patch(
            "foodx_devops_tools.deploy_me._main" ".identify_release_state",
            return_value=TestDeployMe.MockReleaseState.r1,
        )
        with caplog.at_level(logging.DEBUG):
            with split_directories(CLEAN_SPLIT.copy()) as (
                client_config,
                system_config,
            ):
                mock_input = [
                    str(client_config),
                    str(system_config),
                    "--git-ref",
                    "refs/heads/main",
                ]

                result = click_runner.invoke(deploy_me, mock_input)

                assert result.exit_code == 0



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Azure resource utilities."""

import asyncio
import json
import logging
import typing

from foodx_devops_tools.utility import run_async_command

from .model import AzureSubscriptionConfiguration

log = logging.getLogger(__name__)


class ResourceError(Exception):
    """Problem occurred with resource related actions."""


async def list_resources(
    subscription: AzureSubscriptionConfiguration,
    name: typing.Optional[str] = None,
    location: typing.Optional[str] = None,
    tag: typing.Optional[str] = None,
) -> typing.List[dict]:
    """
    List resources in a specified subscription.

    The same name, location and tag glob filters as Azure CLI utility may be
    specified.

    Args:
        subscription: Subscription to list resources in.
        name: Name of resource to search for [optional].
        location: Location of resource to search for [optional].
        tag: Tag(s) of resource(s) to search for [optional].

    Returns:
        List of resource found. Empty list otherwise.
    """
    if (not name) and (not location) and (not tag):
        raise ResourceError(
            "Must specify at least one of name, location or tag"
        )

    this_command = [
        "az",
        "resource",
        "list",
        "--subscription",
        subscription.subscription_id,
    ]
    if name:
        this_command += ["--name", name]
    if location:
        this_command += ["--location", location]
    if tag:
        this_command += ["--tag", tag]
    log.debug("{0}".format(str(this_command)))
    result = await run_async_command(this_command)
    log.debug("list resources stdout, {0}".format(result.out))
    log.debug("list resources stderr, {0}".format(result.error))

    data: typing.List[dict] = json.loads(result.out)
    log.debug("{0}".format(str(data)))
    return data


async def check_exists(
    resource_name: str, subscription: AzureSubscriptionConfiguration
) -> typing.List[dict]:
    """
    Check that the named resource exists.

    Take care that multiple resources of different types could have the same
    name, so a non-null result is not necessarily affirmation of the
    existence you intend.

    Args:
        resource_name: Name of resource to find.
        subscription: Subscription to search for resource.

    Returns:
        Dictionary of resource(s) found. Empty dictionary otherwise.
    Raises:
        ResourceError: If any problems occur doing the check.
    """
    try:
        result_data = await list_resources(subscription, name=resource_name)

        matches = list()
        for this_resource in result_data:
            if this_resource["name"] == resource_name:
                matches.append(this_resource.copy())

        return matches
    except asyncio.CancelledError:
        # should almost always let async cancelled exceptions propagate.
        raise
    except Exception as e:
        raise ResourceError(
            "Problem acquiring resource  data, {0}".format(resource_name)
        ) from e



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

import logging
import pathlib

from foodx_devops_tools.pipeline_config import PipelineConfigurationPaths

from ._exceptions import ConfigurationPathsError

log = logging.getLogger(__name__)

PIPELINE_CONFIG_FILES = {
    "clients.yml",
    "release_states.yml",
    "deployments.yml",
    "frames.yml",
    "subscriptions.yml",
    "systems.yml",
    "tenants.yml",
}


def acquire_configuration_paths(
    client_config: pathlib.Path, system_config: pathlib.Path
) -> PipelineConfigurationPaths:
    """Acquire system, pipeline configuration paths."""
    client_files = [
        x
        for x in client_config.iterdir()
        if x.is_file() and x.name in PIPELINE_CONFIG_FILES
    ]
    system_files = [
        x
        for x in system_config.iterdir()
        if x.is_file() and x.name in PIPELINE_CONFIG_FILES
    ]

    if len(client_files + system_files) > len(PIPELINE_CONFIG_FILES):
        # must be duplicate files between the directories
        log.debug("client files, {0}".format(str(client_files)))
        log.debug("system files, {0}".format(str(system_files)))
        raise ConfigurationPathsError(
            "Duplicate files between "
            "directories, {0}, {1}".format(client_config, system_config)
        )

    path_arguments = {
        x.name.strip(".yml"): x
        for x in set(client_files).union(set(system_files))
    }
    result = PipelineConfigurationPaths(**path_arguments)
    return result



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Declarations relating to deployment management."""

import dataclasses
import typing

T = typing.TypeVar("T", bound="DeploymentState")


@dataclasses.dataclass
class DeploymentState:
    """System deployment tuple."""

    client: str
    release_state: str
    system: str

    def __str__(self: T) -> str:
        """Convert ``DeploymentState`` object to normalized representation."""
        result = "-".join([self.system, self.client, self.release_state])
        return result



#  Copyright (c) 2021 Food-X Technologies
#
#  This file is part of foodx_devops_tools.
#
#  You should have received a copy of the MIT License along with
#  foodx_devops_tools. If not, see <https://opensource.org/licenses/MIT>.

"""Applications, frames configuration I/O."""

import enum
import logging
import pathlib
import typing

import pydantic

from ._loader import load_configuration

log = logging.getLogger(__name__)

ENTITY_NAME = "frames"
ENTITY_SINGULAR = "frame"


class FrameDefinitionsError(Exception):
    """Problem loading frame definitions."""


DependencyDeclarations = typing.List[str]
GlobPathDeclarations = typing.List[str]


@enum.unique
class DeploymentMode(str, enum.Enum):
    """Azure Cloud resource group deployment mode enumeration."""

    complete = "Complete"
    incremental = "Incremental"


class TriggersDefinition(pydantic.BaseModel):
    """Definition of deployment triggers."""

    paths: GlobPathDeclarations


class ApplicationDeploymentDefinition(pydantic.BaseModel):
    """Application resource group deployment definition."""

    resource_group: str
    mode: DeploymentMode

    arm_file: typing.Optional[pathlib.Path]
    puff_file: typing.Optional[pathlib.Path]


ApplicationDeclarations = typing.Dict[
    str, typing.List[ApplicationDeploymentDefinition]
]


class SingularFrameDefinition(pydantic.BaseModel):
    """A single frame definition."""

    applications: ApplicationDeclarations
    folder: pathlib.Path

    depends_on: typing.Optional[DependencyDeclarations]
    triggers: typing.Optional[TriggersDefinition]


class FramesTriggersDefinition(pydantic.BaseModel):
    """Definition of frames and triggers."""

    frames: typing.Dict[str, SingularFrameDefinition]
    triggers: typing.Optional[TriggersDefinition]

    @pydantic.validator(ENTITY_NAME)
    def check_frames(cls: pydantic.BaseModel, frames_candidate: dict) -> dict:
        """Validate ``frames`` field."""
        frame_names: list = list(frames_candidate.keys())
        if any([not x for x in frame_names]):
            message = "Empty {0} names prohibited".format(ENTITY_SINGULAR)
            # log the message here because pydantic exception handling
            # masks the true exception that caused a validation failure.
            log.error(message)
            raise ValueError(message)
        if len(set(frame_names)) != len(frame_names):
            message = "Duplicate {0} names prohibited".format(ENTITY_SINGULAR)
            # log the message here because pydantic exception handling
            # masks the true exception that caused a validation failure.
            log.error(message)
            raise ValueError(message)

        return frames_candidate

    @pydantic.root_validator
    def check_dependencies(
        cls: pydantic.BaseModel, frames_candidate: dict
    ) -> dict:
        """
        Validate frame dependencies.

        Frames must only declare dependencies on other frames.
        """
        frames: dict = frames_candidate.get("frames", dict())
        frame_names: set = set(frames.keys())
        for name, item in frames.items():
            if item.depends_on and (
                any([x not in frame_names for x in item.depends_on])
            ):
                message = "Unknown frame dependency, {0}".format(
                    str(item.depends_on)
                )
                # log the message here because pydantic exception handling
                # masks the true exception that caused a validation failure.
                log.error(message)
                raise ValueError(message)

        return frames_candidate


ValueType = FramesTriggersDefinition

T = typing.TypeVar("T", bound="FramesDefinition")


class FramesDefinition(pydantic.BaseModel):
    """Definition of frames."""

    frames: ValueType


def load_frames(file_path: pathlib.Path) -> FramesDefinition:
    """
    Load frame definitions from file.

    Args:
        file_path: Path to frame definitions file.

    Returns:
        Frame definitions.
    Raises:
        FrameDefinitionsError: If an error occurs loading the file.
    """
    log.info("loading frame definitions from file, {0}".format(file_path))
    result = load_configuration(
        file_path, FramesDefinition, FrameDefinitionsError, ENTITY_NAME
    )

    return typing.cast(FramesDefinition, result)
